"""Griffe Sphinx package.

Parse Sphinx-comments above attributes as docstrings.
"""

from __future__ import annotations

from griffe_sphinx._internals.extension import SphinxCommentsExtension

__all__: list[str] = ["SphinxCommentsExtension"]



"""Generate the code reference pages and navigation."""

from pathlib import Path

import mkdocs_gen_files

nav = mkdocs_gen_files.Nav()
mod_symbol = '<code class="doc-symbol doc-symbol-nav doc-symbol-module"></code>'

root = Path(__file__).parent.parent
src = root / "src"

for path in sorted(src.rglob("*.py")):
    module_path = path.relative_to(src).with_suffix("")
    doc_path = path.relative_to(src).with_suffix(".md")
    full_doc_path = Path("reference", doc_path)

    parts = tuple(module_path.parts)

    if parts[-1] == "__init__":
        parts = parts[:-1]
        doc_path = doc_path.with_name("index.md")
        full_doc_path = full_doc_path.with_name("index.md")
    elif parts[-1].startswith("_"):
        continue

    nav_parts = [f"{mod_symbol} {part}" for part in parts]
    nav[tuple(nav_parts)] = doc_path.as_posix()

    with mkdocs_gen_files.open(full_doc_path, "w") as fd:
        ident = ".".join(parts)
        fd.write(f"---\ntitle: {ident}\n---\n\n::: {ident}")

    mkdocs_gen_files.set_edit_path(full_doc_path, ".." / path.relative_to(root))

with mkdocs_gen_files.open("reference/SUMMARY.md", "w") as nav_file:
    nav_file.writelines(nav.build_literate_nav())



"""Functions related to Insiders funding goals."""

from __future__ import annotations

import json
import logging
import os
import posixpath
from dataclasses import dataclass
from datetime import date, datetime, timedelta
from itertools import chain
from pathlib import Path
from typing import Iterable, cast
from urllib.error import HTTPError
from urllib.parse import urljoin
from urllib.request import urlopen

import yaml

logger = logging.getLogger(f"mkdocs.logs.{__name__}")


def human_readable_amount(amount: int) -> str:  # noqa: D103
    str_amount = str(amount)
    if len(str_amount) >= 4:  # noqa: PLR2004
        return f"{str_amount[:len(str_amount)-3]},{str_amount[-3:]}"
    return str_amount


@dataclass
class Project:
    """Class representing an Insiders project."""

    name: str
    url: str


@dataclass
class Feature:
    """Class representing an Insiders feature."""

    name: str
    ref: str | None
    since: date | None
    project: Project | None

    def url(self, rel_base: str = "..") -> str | None:  # noqa: D102
        if not self.ref:
            return None
        if self.project:
            rel_base = self.project.url
        return posixpath.join(rel_base, self.ref.lstrip("/"))

    def render(self, rel_base: str = "..", *, badge: bool = False) -> None:  # noqa: D102
        new = ""
        if badge:
            recent = self.since and date.today() - self.since <= timedelta(days=60)  # noqa: DTZ011
            if recent:
                ft_date = self.since.strftime("%B %d, %Y")  # type: ignore[union-attr]
                new = f' :material-alert-decagram:{{ .new-feature .vibrate title="Added on {ft_date}" }}'
        project = f"[{self.project.name}]({self.project.url}) — " if self.project else ""
        feature = f"[{self.name}]({self.url(rel_base)})" if self.ref else self.name
        print(f"- [{'x' if self.since else ' '}] {project}{feature}{new}")


@dataclass
class Goal:
    """Class representing an Insiders goal."""

    name: str
    amount: int
    features: list[Feature]
    complete: bool = False

    @property
    def human_readable_amount(self) -> str:  # noqa: D102
        return human_readable_amount(self.amount)

    def render(self, rel_base: str = "..") -> None:  # noqa: D102
        print(f"#### $ {self.human_readable_amount} — {self.name}\n")
        if self.features:
            for feature in self.features:
                feature.render(rel_base)
            print("")
        else:
            print("There are no features in this goal for this project.  ")
            print(
                "[See the features in this goal **for all Insiders projects.**]"
                f"(https://pawamoy.github.io/insiders/#{self.amount}-{self.name.lower().replace(' ', '-')})",
            )


def load_goals(data: str, funding: int = 0, project: Project | None = None) -> dict[int, Goal]:
    """Load goals from JSON data.

    Parameters:
        data: The JSON data.
        funding: The current total funding, per month.
        origin: The origin of the data (URL).

    Returns:
        A dictionaries of goals, keys being their target monthly amount.
    """
    goals_data = yaml.safe_load(data)["goals"]
    return {
        amount: Goal(
            name=goal_data["name"],
            amount=amount,
            complete=funding >= amount,
            features=[
                Feature(
                    name=feature_data["name"],
                    ref=feature_data.get("ref"),
                    since=feature_data.get("since") and datetime.strptime(feature_data["since"], "%Y/%m/%d").date(),  # noqa: DTZ007
                    project=project,
                )
                for feature_data in goal_data["features"]
            ],
        )
        for amount, goal_data in goals_data.items()
    }


def _load_goals_from_disk(path: str, funding: int = 0) -> dict[int, Goal]:
    project_dir = os.getenv("MKDOCS_CONFIG_DIR", ".")
    try:
        data = Path(project_dir, path).read_text()
    except OSError as error:
        raise RuntimeError(f"Could not load data from disk: {path}") from error
    return load_goals(data, funding)


def _load_goals_from_url(source_data: tuple[str, str, str], funding: int = 0) -> dict[int, Goal]:
    project_name, project_url, data_fragment = source_data
    data_url = urljoin(project_url, data_fragment)
    try:
        with urlopen(data_url) as response:  # noqa: S310
            data = response.read()
    except HTTPError as error:
        raise RuntimeError(f"Could not load data from network: {data_url}") from error
    return load_goals(data, funding, project=Project(name=project_name, url=project_url))


def _load_goals(source: str | tuple[str, str, str], funding: int = 0) -> dict[int, Goal]:
    if isinstance(source, str):
        return _load_goals_from_disk(source, funding)
    return _load_goals_from_url(source, funding)


def funding_goals(source: str | list[str | tuple[str, str, str]], funding: int = 0) -> dict[int, Goal]:
    """Load funding goals from a given data source.

    Parameters:
        source: The data source (local file path or URL).
        funding: The current total funding, per month.

    Returns:
        A dictionaries of goals, keys being their target monthly amount.
    """
    if isinstance(source, str):
        return _load_goals_from_disk(source, funding)
    goals = {}
    for src in source:
        source_goals = _load_goals(src, funding)
        for amount, goal in source_goals.items():
            if amount not in goals:
                goals[amount] = goal
            else:
                goals[amount].features.extend(goal.features)
    return {amount: goals[amount] for amount in sorted(goals)}


def feature_list(goals: Iterable[Goal]) -> list[Feature]:
    """Extract feature list from funding goals.

    Parameters:
        goals: A list of funding goals.

    Returns:
        A list of features.
    """
    return list(chain.from_iterable(goal.features for goal in goals))


def load_json(url: str) -> str | list | dict:  # noqa: D103
    with urlopen(url) as response:  # noqa: S310
        return json.loads(response.read().decode())


data_source = globals()["data_source"]
sponsor_url = "https://github.com/sponsors/pawamoy"
data_url = "https://raw.githubusercontent.com/pawamoy/sponsors/main"
numbers: dict[str, int] = load_json(f"{data_url}/numbers.json")  # type: ignore[assignment]
sponsors: list[dict] = load_json(f"{data_url}/sponsors.json")  # type: ignore[assignment]
current_funding = numbers["total"]
sponsors_count = numbers["count"]
goals = funding_goals(data_source, funding=current_funding)
ongoing_goals = [goal for goal in goals.values() if not goal.complete]
unreleased_features = sorted(
    (ft for ft in feature_list(ongoing_goals) if ft.since),
    key=lambda ft: cast(date, ft.since),
    reverse=True,
)



"""Debugging utilities."""

from __future__ import annotations

import os
import platform
import sys
from dataclasses import dataclass
from importlib import metadata


@dataclass
class Variable:
    """Dataclass describing an environment variable."""

    name: str
    """Variable name."""
    value: str
    """Variable value."""


@dataclass
class Package:
    """Dataclass describing a Python package."""

    name: str
    """Package name."""
    version: str
    """Package version."""


@dataclass
class Environment:
    """Dataclass to store environment information."""

    interpreter_name: str
    """Python interpreter name."""
    interpreter_version: str
    """Python interpreter version."""
    interpreter_path: str
    """Path to Python executable."""
    platform: str
    """Operating System."""
    packages: list[Package]
    """Installed packages."""
    variables: list[Variable]
    """Environment variables."""


def _interpreter_name_version() -> tuple[str, str]:
    if hasattr(sys, "implementation"):
        impl = sys.implementation.version
        version = f"{impl.major}.{impl.minor}.{impl.micro}"
        kind = impl.releaselevel
        if kind != "final":
            version += kind[0] + str(impl.serial)
        return sys.implementation.name, version
    return "", "0.0.0"


def get_version(dist: str = "griffe-sphinx") -> str:
    """Get version of the given distribution.

    Parameters:
        dist: A distribution name.

    Returns:
        A version number.
    """
    try:
        return metadata.version(dist)
    except metadata.PackageNotFoundError:
        return "0.0.0"


def get_debug_info() -> Environment:
    """Get debug/environment information.

    Returns:
        Environment information.
    """
    py_name, py_version = _interpreter_name_version()
    packages = ["griffe-sphinx"]
    variables = ["PYTHONPATH", *[var for var in os.environ if var.startswith("GRIFFE_SPHINX")]]
    return Environment(
        interpreter_name=py_name,
        interpreter_version=py_version,
        interpreter_path=sys.executable,
        platform=platform.platform(),
        variables=[Variable(var, val) for var in variables if (val := os.getenv(var))],
        packages=[Package(pkg, get_version(pkg)) for pkg in packages],
    )


def print_debug_info() -> None:
    """Print debug/environment information."""
    info = get_debug_info()
    print(f"- __System__: {info.platform}")
    print(f"- __Python__: {info.interpreter_name} {info.interpreter_version} ({info.interpreter_path})")
    print("- __Environment variables__:")
    for var in info.variables:
        print(f"  - `{var.name}`: `{var.value}`")
    print("- __Installed packages__:")
    for pkg in info.packages:
        print(f"  - `{pkg.name}` v{pkg.version}")


if __name__ == "__main__":
    print_debug_info()
